package com.sillycycle.bagleyd.abacus;

/*
 * @(#)AbacusFrame.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.security.AccessControlException;
import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.util.EnumSet;
import java.util.Random;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.ButtonGroup;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import com.sillycycle.bagleyd.abacus.learn.AbacusTest;
import com.sillycycle.bagleyd.abacus.learn.DemoDialog;
import com.sillycycle.bagleyd.abacus.learn.TeachDialog;
import com.sillycycle.bagleyd.abacus.learn.TestDialog;
import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.Colour;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.PartialDisableComboBox;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Slider;
import com.sillycycle.bagleyd.util.Spinner;
import com.sillycycle.bagleyd.util.TextPrompt;

/**
 * The <code>AbacusFrame</code> class is to be run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 *
 * @author	https://www.sillycycle.com/abacus.html
 */

public class AbacusFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_RAILS = 24; //99
	static final int MAX_DELAY = 500;
	// this seems best way but still ugly
	/*static Map<String, Character> COMMAND_MAP =
			new HashMap<String, Character>() {{
				put("Clear", 'c'); put("Complement", '-');
				put("Undo", 'u'); put("Redo", 'r');
				put("Increment", 'i'); put("Decrement", 'd');
				put("Format", 'f');
				put("Roman Nvmerals", 'v');
				put("Latin ~", '~');
				put("Group", 'g');
				put("Place On Rail", '#');
				put("Vertical", '|');
				put("Sign", 's');
				put("Quarter", 'p');
				put("Quarter Percent", 't');
				...
		}
	};*/
	static final String[] COMMAND_STRINGS = new String[] {
		"Clear", "Complement -", "Undo", "Redo",
		"Increment", "Decrement", "Format",
		"Roman Nvmerals", "Latin ~",
		"Group", "Place On Rail #", "Vertical |",
		"Sign", "Piece", "Piece Percent",
		"Subdeck", "Museum", "Roman Markerz",
		"Subdecks Separated", "Alt Subbead Placement",
		"Anomaly",
		"Right To Left Add +", "Right To Left Mult *",
		"Sound @", "Speed Up >", "Slow Down <",
		"Demo", "Teach $",
		"Description ?", "Features !", "References ^", "About",
		"Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'c', '-', 'u', 'r',
		'i', 'd', 'f',
		'v', '~',
		'g', '#', '|',
		's', 'p', 't',
		'b', 'm', 'z',
		'k', 'l',
		'y',
		'+', '*',
		'@', '>', '<',
		'o', '$',
		'?', '!', '^', 'a',
		'x'};

	static ArgumentParser argumentParser = null;
	static DocumentBuilderFactory docBuilderFactory = null;
	static DocumentBuilder docBuilder = null;
	static Document doc = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 400;
	int heightValue = 400;
	static int xValue = 0;
	static int yValue = 0;
	Color fgColor = Colour.tan;
	Color bgColor = Colour.steelBlue;
	Color borderColor = Colour.gray25;
	Color primaryBeadColor = Colour.darkRed;
	Color leftAuxBeadColor = Colour.limeGreen;
	Color rightAuxBeadColor = Colour.gainsboro;
	Color secondaryBeadColor = Colour.brown;
	Color highlightBeadColor = Colour.magenta;
	Color primaryMarkerColor, leftAuxMarkerColor, rightAuxMarkerColor;
	Color primaryRailColor = Colour.gold;
	Color secondaryRailColor = Colour.silver;
	Color highlightRailColor = Colour.purple;
	Color lineRailColor = Colour.black;
	boolean sound = false;
	int delayValue = 50; // msec
	int inc = 10;
	int testValue = 0; // minutes
	boolean script = false;
	boolean controlValue = true;
	boolean demoValue = false;
	boolean rightToLeftAddValue = false, rightToLeftMultValue = false;
	boolean leeValue = false;
	int railsValue = AbacusInterface.DEFAULT_RAILS;
	int leftAuxRailsValue = AbacusInterface.DEFAULT_LEFT_AUX_RAILS;
	int rightAuxRailsValue = AbacusInterface.DEFAULT_RIGHT_AUX_RAILS;
	int totalAuxRailsValue = leftAuxRailsValue + rightAuxRailsValue;
	boolean verticalValue = false;
	int colorSchemeValue = 0;
	boolean slotValue = false;
	boolean diamondValue = false;
	int railIndexValue = 0;
	boolean topOrientValue = AbacusInterface.DEFAULT_TOP_ORIENT;
	boolean bottomOrientValue = AbacusInterface.DEFAULT_BOTTOM_ORIENT;
	int topNumberValue = AbacusInterface.DEFAULT_TOP_NUMBER;
	int bottomNumberValue = AbacusInterface.DEFAULT_BOTTOM_NUMBER;
	int topFactorValue = AbacusInterface.DEFAULT_TOP_FACTOR;
	int bottomFactorValue = AbacusInterface.DEFAULT_BOTTOM_FACTOR;
	int topSpacesValue = AbacusInterface.DEFAULT_TOP_SPACES;
	int bottomSpacesValue = AbacusInterface.DEFAULT_BOTTOM_SPACES;
	int topPieceValue = 0;
	int bottomPieceValue = 0;
	int topPiecePercentValue = 0;
	int bottomPiecePercentValue = 0;
	int shiftPercentValue = 2;
	int subdeckValue = AbacusInterface.DEFAULT_SUBDECKS;
	int subbeadValue = AbacusInterface.DEFAULT_SUBBEADS;
	boolean signValue = false;
	int decimalPositionValue = 2; /* Good default for most currencies */
	boolean groupValue = false;
	int groupSizeValue = AbacusInterface.DEFAULT_GROUP_SIZE;
	boolean placeOnRailValue = false;
	boolean decimalCommaValue = false;
	int baseValue = AbacusInterface.DEFAULT_BASE;
	int subbaseValue = 0;
	int anomalyValue = 0;
	int shiftAnomalyValue = 2;
	int anomalySqValue = 0;
	int shiftAnomalySqValue = 2;
	int displayBaseValue = AbacusInterface.DEFAULT_BASE;
	boolean pressOffsetValue = false;
	AbacusInterface.RomanNumeralsFormat romanNumeralsValue =
			AbacusInterface.RomanNumeralsFormat.NONE;
	boolean latinValue = false;
	AbacusInterface.RomanMarkersFormat romanMarkersValue =
			AbacusInterface.RomanMarkersFormat.NONE;
	int modeValue = AbacusInterface.Modes.CHINESE.ordinal();
	int museumValue = -1;
	boolean subdecksSeparatedValue = true;
	boolean altSubbeadPlacementValue = false;
	Random generator;
	int valueTextLength = 42 - 14; /* Roman Numerals fit */
	static boolean debug = false;
	//static boolean debug = true;
	JPanel topPanel = new JPanel();
	PartialDisableComboBox modeComboBox = new PartialDisableComboBox();
	Abacus abacus = null, leftAuxAbacus = null, rightAuxAbacus = null;
	transient AbacusCalc abacusCalc = new AbacusCalc(abacus);
	transient AbacusCalc leftAuxAbacusCalc = new AbacusCalc(leftAuxAbacus);
	transient AbacusCalc rightAuxAbacusCalc = new AbacusCalc(rightAuxAbacus);
	JTextField leftAuxValueTextField = new JTextField("");
	JTextField rightAuxValueTextField = new JTextField("");
	JTextField valueTextField = new JTextField("");
	// To get multicolor in a single field, need documentListener
	//JEditorPane leftAuxValueTextField = new JEditorPane();
	//JEditorPane rightAuxValueTextField = new JEditorPane();
	//JEditorPane valueTextField = new JEditorPane();
	TextPrompt leftAuxValueTextPrompt = null;
	TextPrompt rightAuxValueTextPrompt = null;
	TextPrompt valueTextPrompt = null;
	OKDialog aDialog = null;
	DemoDialog demoDialog = null;
	TestDialog testDialog = null;
	TeachDialog teachDialog = null;
	StringBuilder stringBuilderSave;
	boolean mathDone = true;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	boolean isPopped = false;
	Font font = new Font("Verdana", Font.PLAIN, 12);
	JMenu secondaryRailsMenu, museumMenu;
	JMenu pieceRailMenu, piecePercentRailMenu, subdeckRailsMenu;
	JMenu romanNumeralsMenu, romanMarkersMenu;
	JMenuItem[] romanNumeralsMenuItem = null;
	JCheckBoxMenuItem latinMenuItem = null;
	JCheckBoxMenuItem groupMenuItem = null;
	JCheckBoxMenuItem placeOnRailMenuItem = null;
	JCheckBoxMenuItem verticalMenuItem = null;
	JCheckBoxMenuItem signMenuItem = null;
	JMenuItem[] pieceMenuItem = null;
	JMenuItem[] piecePercentMenuItem = null;
	JMenuItem[] subdeckMenuItem = null;
	JCheckBoxMenuItem subdecksSeparatedMenuItem = null;
	JCheckBoxMenuItem altSubbeadPlacementMenuItem = null;
	JMenuItem[] museumMenuItem = null;
	JMenuItem[] romanMarkersMenuItem = null;
	JMenuItem[] anomalyMenuItem = null;
	JCheckBoxMenuItem rightToLeftAddMenuItem = null;
	JCheckBoxMenuItem rightToLeftMultMenuItem = null;
	JCheckBoxMenuItem soundMenuItem = null;
	//Slider railChanger = null;
	Slider displayChanger = null, baseChanger = null;
	Slider delayChanger = null;
	Spinner railChanger = null;
	//Spinner displayChanger = null, baseChanger = null;
	//Spinner delayChanger = null;

	public String getDemoString() {
		return this.stringBuilderSave.toString();
	}

	public boolean getDemo() {
		return demoValue;
	}

	public boolean getTest() {
		return (testValue != 0);
	}

	public boolean getTeach() {
		return (teachDialog != null);
	}

	public boolean getRightToLeftAdd() {
		return rightToLeftAddValue;
	}

	public void setRightToLeftAdd(boolean rightToLeftAdd) {
		rightToLeftAddValue = rightToLeftAdd;
	}

	public boolean getRightToLeftMult() {
		return rightToLeftMultValue;
	}

	public void setRightToLeftMult(boolean rightToLeftMult) {
		rightToLeftMultValue = rightToLeftMult;
	}

	public boolean getLee() {
		return leeValue;
	}

	public Abacus getAbacus(int aux) {
		if (leeValue && aux == AbacusInterface.LEFT_AUX)
			return leftAuxAbacus;
		else if (leeValue && aux == AbacusInterface.RIGHT_AUX)
			return rightAuxAbacus;
		else /*if (aux == AbacusInterface.PRIMARY)*/
			return abacus;
	}

	public boolean mainAbacus(AbacusCanvas ab) {
		return abacus.equals(ab);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void initSound() {
		/* There is a noticeable delay on initializing, so maybe it should
		 *  be done first.  On the other hand, it may hang which would
		 *  be worse.
		 */
	}

	public void toggleSound() {
		sound = !sound;
		if (controlValue)
			soundMenuItem.setSelected(sound);
	}

	public boolean getScript() {
		return script;
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public AbacusInterface.RomanNumeralsFormat getOpt(String param,
			AbacusInterface.RomanNumeralsFormat value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			AbacusInterface.RomanNumeralsFormat[] vals =
					AbacusInterface.RomanNumeralsFormat.values();
			for (int i = 0; i < vals.length; i++) {
				if (string.equalsIgnoreCase(vals[i].toString())) {
					 return vals[i];
				}
			}
			return value;
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public AbacusInterface.RomanMarkersFormat getOpt(String param,
			AbacusInterface.RomanMarkersFormat value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			AbacusInterface.RomanMarkersFormat[] vals =
					AbacusInterface.RomanMarkersFormat.values();
			for (int i = 0; i < vals.length; i++) {
				if (string.equalsIgnoreCase(vals[i].toString())) {
					 return vals[i];
				}
			}
			return value;
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		String string;

		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		borderColor = getOpt("borderColor", borderColor);
		primaryBeadColor = getOpt("primaryBeadColor", primaryBeadColor);
		leftAuxBeadColor = getOpt("leftAuxBeadColor",
			leftAuxBeadColor);
		rightAuxBeadColor = getOpt("rightAuxBeadColor",
			rightAuxBeadColor);
		secondaryBeadColor = getOpt("secondaryBeadColor",
			secondaryBeadColor);
		highlightBeadColor = getOpt("highlightBeadColor",
			highlightBeadColor);
		primaryMarkerColor = getOpt("primaryMarkerColor",
			primaryBeadColor);
		leftAuxMarkerColor = getOpt("leftAuxMarkerColor",
			leftAuxBeadColor);
		rightAuxMarkerColor = getOpt("rightAuxMarkerColor",
			rightAuxBeadColor);
		primaryRailColor = getOpt("primaryRailColor",
			primaryRailColor);
		secondaryRailColor = getOpt("secondaryRailColor",
			secondaryRailColor);
		highlightRailColor = getOpt("highlightRailColor",
			highlightRailColor);
		lineRailColor = getOpt("lineRailColor",
			lineRailColor);
		sound = getOpt("sound", sound);
		script = getOpt("script", script);
		controlValue = getOpt("control", controlValue);
		rightToLeftAddValue = getOpt("rightToLeftAdd",
			rightToLeftAddValue);
		rightToLeftMultValue = getOpt("rightToLeftMult",
			rightToLeftMultValue);
		leeValue = getOpt("lee", leeValue);
		railsValue = getOpt("rails", railsValue);
		if (railsValue < AbacusInterface.MIN_RAILS)
			railsValue = AbacusInterface.DEFAULT_RAILS;
		leftAuxRailsValue = getOpt("leftAuxRails", leftAuxRailsValue);
		if (leftAuxRailsValue < AbacusInterface.MIN_RAILS)
			leftAuxRailsValue =
				AbacusInterface.DEFAULT_LEFT_AUX_RAILS;
		rightAuxRailsValue = getOpt("rightAuxRails",
			rightAuxRailsValue);
		if (rightAuxRailsValue < AbacusInterface.MIN_RAILS)
			rightAuxRailsValue =
				AbacusInterface.DEFAULT_RIGHT_AUX_RAILS;
		totalAuxRailsValue = leftAuxRailsValue + rightAuxRailsValue;
		colorSchemeValue = getOpt("colorScheme", colorSchemeValue);
		slotValue = getOpt("slot", slotValue);
		diamondValue = getOpt("diamond", diamondValue);
		railIndexValue = getOpt("railIndex", railIndexValue);
		topOrientValue = getOpt("topOrient", topOrientValue);
		bottomOrientValue = getOpt("bottomOrient", bottomOrientValue);
		topNumberValue = getOpt("topNumber", topNumberValue);
		bottomNumberValue = getOpt("bottomNumber", bottomNumberValue);
		topFactorValue = getOpt("topFactor", topFactorValue);
		bottomFactorValue = getOpt("bottomFactor", bottomFactorValue);
		topSpacesValue = getOpt("topSpaces", topSpacesValue);
		bottomSpacesValue = getOpt("bottomSpaces", bottomSpacesValue);
		topPieceValue = getOpt("topPiece", topPieceValue);
		bottomPieceValue = getOpt("bottomPiece", bottomPieceValue);
		topPiecePercentValue = getOpt("topPiecePercent",
			topPiecePercentValue);
		bottomPiecePercentValue = getOpt("bottomPiecePercent",
			bottomPiecePercentValue);
		shiftPercentValue = getOpt("shiftPercent", shiftPercentValue);
		subdeckValue = getOpt("subdeck", subdeckValue);
		subbeadValue = getOpt("subbead", subbeadValue);
		signValue = getOpt("sign", signValue);
		decimalPositionValue = getOpt("decimalPosition",
			decimalPositionValue);
		if (bottomPieceValue != 0) // kludge
			decimalPositionValue++;
		if (bottomPiecePercentValue != 0) // kludge
			decimalPositionValue++;
		groupValue = getOpt("group", groupValue);
		groupSizeValue = getOpt("groupSize", groupSizeValue);
		placeOnRailValue = getOpt("placeOnRail", placeOnRailValue);
		decimalCommaValue = getOpt("decimalComma", decimalCommaValue);
		baseValue = getOpt("base", baseValue);
		if (baseValue < AbacusInterface.MIN_BASE ||
				baseValue > AbacusInterface.MAX_BASE)
			baseValue = AbacusInterface.DEFAULT_BASE;
		subbaseValue = getOpt("subbase", subbaseValue);
		subdecksSeparatedValue = getOpt("subdecksSeparated", subdecksSeparatedValue);
		altSubbeadPlacementValue = getOpt("altSubbeadPlacement", altSubbeadPlacementValue);
		anomalyValue = getOpt("anomaly", anomalyValue);
		shiftAnomalyValue = getOpt("shiftAnomaly", shiftAnomalyValue);
		anomalySqValue = getOpt("anomalySq", anomalySqValue);
		shiftAnomalySqValue = getOpt("shiftAnomalySq",
			shiftAnomalySqValue);
		displayBaseValue = getOpt("displayBase", displayBaseValue);
		if (displayBaseValue < AbacusInterface.MIN_BASE ||
				displayBaseValue > AbacusInterface.MAX_BASE)
			displayBaseValue = AbacusInterface.DEFAULT_BASE;
		pressOffsetValue = getOpt("pressOffset", pressOffsetValue);
		romanNumeralsValue = getOpt("romanNumerals",
			romanNumeralsValue);
		romanMarkersValue = getOpt("romanMarkers", romanMarkersValue);
		latinValue = getOpt("latin", latinValue);
		string = getOpt("format");
		if (string != null) {
			modeValue = AbacusInterface.setModeFromFormat(string);
		}
		modeValue = getOpt("mode", modeValue);
		if (modeValue < AbacusInterface.Modes.CHINESE.ordinal() ||
				modeValue > AbacusInterface.Modes.GENERIC.ordinal())
			modeValue = AbacusInterface.Modes.CHINESE.ordinal();
		verticalValue = (modeValue == AbacusInterface.Modes.RUSSIAN.ordinal() ||
			modeValue == AbacusInterface.Modes.DANISH.ordinal() ||
			modeValue == AbacusInterface.Modes.MEDIEVAL.ordinal());
		verticalValue = getOpt("vertical", verticalValue);
		string = getOpt("museum");
		if (string != null) {
			museumValue = AbacusInterface.setMuseumFromFormat(string);
		}
		museumValue = getOpt("submode", museumValue);
		if (museumValue < 0 ||
				museumValue >= AbacusInterface.Museum.values().length)
			museumValue = generator.nextInt(AbacusInterface.Museum.values().length);
		delayValue = getOpt("delay", delayValue);
		testValue = getOpt("test", testValue);
	}


	public void createDemoDialog(String msg) {
		if (demoDialog == null) {
//			FontMetrics fm = getFontMetrics(this.getFont());
			FontMetrics fm = getFontMetrics(font);

			demoDialog = new DemoDialog(ComponentUtil.findJFrame(this),
				this, AbacusInterface.TITLE + ": Demo",
				msg + "                               ",
				largeIcon);
			demoDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2,
				425, 8 * fm.getHeight());
			demoDialog.setBackground(Color.white);
			demoDialog.setVisible(true);
		}
	}

	public void killDemoDialog() {
		if (demoDialog != null) {
			demoDialog.dispose();
			demoDialog = null;
		}
	}

	public void showDemoMsg(String msg) {
		if (demoDialog == null) {
			createDemoDialog(msg);
		} else {
			demoDialog.showDemoMessage(msg);
		}
	}

	public void showDemoMsg2(String msg0, String msg1, int delay) {
		demoDialog.showDemoMessage2(msg0, msg1, delay);
	}

	/*public static String readUserDemo() {
		return "";
	}*/

	public void setTest(String[] testNames, AbacusTest test) {
		if (testDialog == null) {
			createTestDialog(testNames, test);
		}
	}

	public void setTestLabel(String msg) {
		testDialog.showTestLabel(msg);
	}

	public void setTestQuestion(String msg) {
		testDialog.showTestQuestion(msg);
	}

	public void setTestPercentValue(String msg) {
		testDialog.showTestPercentValue(msg);
	}

	public void setGrade(int numberCorrect, int numberQuestions, int grade, String gradeString) {
		testDialog.showTestGrade(numberCorrect, numberQuestions, grade, gradeString);
	}

	public void createTestDialog(String[] testNames, AbacusTest test) {
		if (testDialog == null) {
//			FontMetrics fm = getFontMetrics(this.getFont());
			FontMetrics fm = getFontMetrics(font);

			testDialog = new TestDialog(ComponentUtil.findJFrame(this),
				this, test, AbacusInterface.TITLE + ": Test",
				testNames, largeIcon, testValue);
			testDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2,
				425, 13 * fm.getHeight());
			testDialog.setBackground(Color.white);
			testDialog.setVisible(true);
		}
	}

	public void killTestDialog() {
		if (testDialog != null) {
			testDialog.dispose();
			testDialog = null;
		}
	}

	public void createTeachDialog() {
		if (teachDialog == null) {
//			FontMetrics fm = getFontMetrics(this.getFont());
			FontMetrics fm = getFontMetrics(font);

			teachDialog = new TeachDialog(ComponentUtil.findJFrame(this),
				this, abacus, AbacusInterface.TITLE + ": Teach",
				AbacusInterface.TEACH_STRING0, 0, largeIcon);
			teachDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2,
				600, 8 * fm.getHeight());
			teachDialog.setBackground(Color.white);
			teachDialog.setVisible(true);
		}
	}

	public void showTeachMessage(String msg, int line) {
		teachDialog.showTeachMessage(msg, line);
	}

	public void killTeachDialog() {
		if (teachDialog != null) {
			teachDialog.dispose();
			teachDialog = null;
		}
	}

	public void setTeachMsg(String m, int line) {
		showTeachMessage(m, line);
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				AbacusInterface.TITLE, msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				AbacusInterface.TITLE + ": Description",
"This is an implementation of the classic Chinese Abacus " +
"(Suanpan) which has its origins in the 12th century." + NL +
" " + NL +
"The device has two decks. Each deck, separated by a " +
"partition, normally has 13 rails on which are mounted beads." + NL +
"Each rail on the top deck contains 1 or 2 beads, and each " +
"rod on the bottom deck contains 4 or 5 beads.  Each bead on" + NL +
"the upper deck has a value of five, while each bead on the " +
"lower deck has value of one. Beads are considered counted," + NL +
"when moved towards the partition separating the decks, i.e. " +
"to add a value of one, a bead in the bottom deck is moved" + NL +
"up, and to add a value of 5, a bead in the top deck is moved " +
"down." + NL +
" " + NL +
"The basic operations of the abacus are addition and subtraction. " +
" Multiplication can be done by mentally multiplying the" + NL +
"digits and adding up the intermediate results on the abacus. " +
" Division would be similar where the intermediate results" + NL +
"are subtracted.  There are techniques like using your thumb " +
"with forefinger which does not apply with mouse entry." + NL +
"Also with multiplication, one can carry out calculations on " +
"different parts of the abacus for scratch work, here it is" + NL +
"nice to have a long abacus." + NL +
" " + NL +
"The pre-WWII Japanese Abacus (Soroban) (or Korean Jupan) " +
"is similar to the Chinese Abacus but has only one bead" + NL +
"per rail on the top deck.  The later Japanese Abacus was " +
"further simplified to have only 4 beads per rail on the" + NL +
"bottom deck." + NL +
" " + NL +
"The Russian Abacus was invented in the 17th century, here " +
"the beads are moved from right to left.  It has colored" + NL +
"beads in the middle for ease of use.  Quarters represent " +
"1/4 Rubles and are only present historically on the Russian" + NL +
"Abacus (Schoty).  Some of the older Schoty have a extra place " +
"for the 1/4 Kopek (quarter percent) as well as the 1/4" + NL +
"Ruble (quarter)." + NL +
" " + NL +
"The Danish Abacus was used in the early 20th century in " +
"elementary schools as a teaching aid." + NL +
" " + NL +
"The Roman Hand-Abacus predates the Chinese Abacus and is " +
"very similar to the later Japanese Abacus, but seems to" + NL +
"have fallen out of use with the Fall of the Roman Empire " +
"(at least 3 are in existence).  The Roman Abaci are brass" + NL +
"plates where the beads move in slots.  In addition to the " +
"normal 7 columns of beads, they generally have 2 special " + NL +
"columns on the right side.  In two examples: the first " +
"special column was for 12ths (12 uncia (ounces) = 1 as) and" + NL +
"had one extra bead in the bottom deck.  Also the last column " +
"was a combination of halves, quarters, and twelfths of an" + NL +
"ounce and had no beads in the top deck and 4 beads at the " +
"bottom (beads did not have to come to the top to be" + NL +
"counted but at one of 3 marked points, where the top bead " +
"was for halves, the next bead for quarters, and the last two" + NL +
"beads for twelfths).  In another surviving example: the 2 " +
"special columns were switched and the combination column" + NL +
"was broken into 3 separate slots.  If available, decimal " +
"input is ignored." + NL +
" " + NL +
"The Medieval Counter is a primitive form of the abacus and " +
"was used in Europe as late as the 1600s.  It was useful" + NL +
"considering they were using it with Roman Numerals.   This " +
"is similar to the Salamis Greek Tablet from 4th or 5th" + NL +
"Century BCE." + NL +
" " + NL +
"The Mesoamerican Nepohualtzintzin is a Japanese Abacus " +
"base 20.  The Mesoamericans had base 20 with the" + NL +
"exception of the 3rd decimal place where instead of " +
"20 * 20 = 400 the third place marked 360 and the 4th place was" + NL +
"20 * 360, etc.  They independently created their own zero " +
"(only Babylon (base 60) and India (base 10) have done this)" + NL +
"but the anomaly took away its true power." + NL +
" " + NL +
"An easy way of figuring out time in seconds given hours, " +
"minutes, and seconds, can be done on the abacus with" + NL +
"Sumerian abacus with topFactor 10, topNumber 5 and " +
"bottomNumber 9 using base 60.  No written record of such a" + NL +
"device exits but its plausible that one did." + NL +
" " + NL +
"The Chinese Solid-and-Broken-Bar System is a base 12 " +
"numbering system and not really an abacus.  When the" + NL +
"abacus is setup in this way though (topFactor 3, topNumber 3, " +
"bottomNumber 2, base 12, displayBase 12), it is easy to" + NL +
"relate the two." + NL +
" " + NL +
"The signed bead is an invention of the author and is not " +
"present on any historical abacus (to his knowledge) and is" + NL +
"used to represent negatives.  \"New & Improved\" abacus " +
"models have two auxiliary decks stacked above the principal" + NL +
"deck that enable multiplication, division, square-root, and " +
"cube-root computations to be performed with equal ease as" + NL +
"addition and subtraction (well, so I have read).",
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				AbacusInterface.TITLE + ": Features",
"Click \"mouse-left\" button on a bead you want to move.  The " +
"beads will shift themselves to vacate the area of the" + NL +
"column that was clicked." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"C\" or \"c\" keys, " +
"to clear the abacus." + NL +
" " + NL +
"Press \"-\" or \"_\" keys to complement the beads on the rails." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo last move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo the last undo." + NL +
" " + NL +
"Press \"I\" or \"i\" keys to increment the number of rails." + NL +
"Press \"D\" or \"d\" keys to decrement the number of rails." + NL +
" " + NL +
"Press \"F\" or \"f\" keys to switch between Chinese, Japanese, " +
"Korean, Russian, Danish, Roman, and Medieval formats." + NL +
"There is an extra \"Generic\" format, this allows one to " +
"break some rules binding the other formats (for example, if one" + NL +
"wanted more beads on top deck than on bottom deck you " +
"would use this, in addition to resource option changes)." + NL +
" " + NL +
"Press \"V\" or \"v\" keys to toggle the Roman Nvmerals " +
"format.  (Pardon humor/typo but ran out of letters)." + NL +
" " + NL +
"Press \"~\" key to toggle Latin Numerals (when Roman " +
"Nvmerals and quarter beads or twelfth beads are activated)." + NL +
" " + NL +
"Press \"G\" or \"g\" keys to toggle (usu.) commas between " +
"groups of digits." + NL +
" " + NL +
"Press \"#\" key to toggle place on rail (or between rails) " +
"for group and decimal separators." + NL +
" " + NL +
"Press \"|\" key to toggle vertical placement (best for " +
"Russian and Danish Abacus, and Medieval Counters)." + NL +
" " + NL +
"Press \"S\" or \"s\" keys to toggle the sign sign bead." + NL +
" " + NL +
"Press \"P\" or \"p\" keys to change the piece setting " +
"from none|quarter|eighth|twelfth beads.  Originally " + NL +
"intended for the Roman Hand Abacus (twelfths) and the " +
"Russian Abacus (quarter rubles)." + NL +
" " + NL +
"Press \"T\" or \"t\" keys to toggle the piece percent " +
"setting for none|quarter|eighth|twelfth beads.  Originally " + NL +
"intended for the older Russian Abacus (quarter kopeks)." + NL + 
" " + NL +
"Press \"B\" or \"b\" keys to change subdeck setting " +
"to be interpretted as none|quarter|eighth|twelfth beads." + NL +
"Originally intended for Roman Hand Abacus twelfths " +
"and possibly eighths.  For twelfths, the the lowest" + NL +
"value of two at bottom of the rightmost column of beads " +
"are a twelfth of the column second from right.  Middle" + NL +
"is worth a quarter and the top is worth a half.  For " +
"eighths, middle and bottom are both worth an eight and" + NL +
"the top a half.  (For quarters, all are worth a quarter.)" + NL +
" " + NL +
"Press \"K\" or \"k\" to toggle the subdeck separation. " + NL +
" " + NL +
"Press \"L\" or \"l\" to toggle which side the subdeck " +
"symbols are placed." + NL +
" " + NL +
"Press \"M\" or \"m\" keys to switch between it, uk, and fr " +
"museum formats." + NL +
" " + NL +
"Press \"Z\" or \"z\" keys to change the Roman markerz " +
"on the frame." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of beads." + NL +
"Press \"<\" or \",\" keys to slow down this movement." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to toggle the demo mode." + NL +
" " + NL +
"Press \"$\" key to toggle the teach mode." + NL +
" " + NL +
"In teach mode, \"+\" key toggles starting side to sum, " +
"\"*\" key toggles for starting side for multiplicand." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program." + NL +
" " + NL +
"The abacus may be resized.  Beads will reshape depending " +
"on the room they have." + NL +
" " + NL +
"Demo Mode: in this mode, the abacus is controlled by the " +
"program.  When started with the demo option, a second" + NL +
"window is presented that should be placed directly below the " +
"abacus-window. Descriptive text, and user prompts are" + NL +
"displayed in this window.  Pressing 'q' during the demo " +
"will quit it.  Clicking the left mouse-button with the" + NL +
"pointer in the window will restart the demo (beginning " +
"of current lesson).",
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				AbacusInterface.TITLE + ": References",
"Luis Fernandes  http://www.ee.ryerson.ca/~elf/abacus/" + NL +
"Lee Kai-chen, How to Learn Lee's Abacus, 1958, 58 pages." + NL +
"Abacus Guide Book, 57 pages." + NL +
"Georges Ifrah, The Universal history of Numbers, Wiley Press " +
"2000, pp 209-211, 288-294." + NL +
"Review of above: http://www.ams.org/notices/200201/rev-dauben.pdf" + NL +
"David Eugene Smith, History of Mathematics Volume II, " +
"Dover Publications, Inc 1958, pp 156-195." + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				AbacusInterface.TITLE + ": About",
"Java AbacusFrame Version 8.6.2" + NL +
"Send bugs (reports or fixes) to the author: " +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/abacus.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void processKey(char key) {
		abacus.processKey(key, false);
	}

	public void setRailChangeListenerValue(int n) {
		if (railChanger != null) {
			railChanger.setValue(n);
		}
	}

	public void setDelayChangeListenerValue(int n) {
		if (delayChanger != null) {
			delayChanger.setValue(n);
		}
	}

	public String getValue() {
		//replace with spaces then trim!!
		return valueTextField.getText().replace(',', ' ').trim();
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = AbacusFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = AbacusFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playBumpAudio() {
		playClip(AbacusInterface.SOUNDS[AbacusInterface.Sounds.BUMP.ordinal()]);
	}

	public void playDripAudio() {
		playClip(AbacusInterface.SOUNDS[AbacusInterface.Sounds.DRIP.ordinal()]);
	}

	public void playMoveAudio() {
		playClip(AbacusInterface.SOUNDS[AbacusInterface.Sounds.MOVE.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (abacus == null)
			return;
		if (leeValue) {
			int newWidth = leftAuxRailsValue * width /
				totalAuxRailsValue;
			int newHeight = height / 4;

			leftAuxAbacus.setBounds(0, 0,
				newWidth, newHeight);
			rightAuxAbacus.setBounds(0, 0,
				width - newWidth, newHeight);
			abacus.setBounds(0, 0,
				width, height - newHeight);
			validate();
		} else {
			abacus.setBounds(0, 0, width, height);
		}
		//repaint(); // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
				secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	private static boolean enableLatin(boolean romanNumerals, int bottomPiece) {
		return (romanNumerals && bottomPiece != 0);
	}

	static boolean checkSecondary(int bottomPiece, int bottomPiecePercent, int decimalPosition,
			boolean slot) {
		if (bottomPiece == 0) {
			System.out.println("bottomPiece = "
				+ bottomPiece);
			return false;
		}
		if (bottomPiecePercent != 0) {
			System.out.println("bottomPiecePercent = "
				+ bottomPiecePercent);
			return false;
		}
		if (!slot) {
			System.out.println("slot = " + slot);
			return false;
		}
		if (decimalPosition != 2) {
			System.out.println("decimalPosition = "
				+ decimalPosition);
			return false;
		}
		return true;
	}

	static void checkSubbase(int bottomPiece, int bottomPiecePercent, int decimalPosition,
			boolean slot, int subbase) {
		if (!checkSecondary(bottomPiece, bottomPiecePercent, decimalPosition, slot))
			return;
		System.out.println("check subbase = " + subbase);
	}

	static boolean enablePiecePercent(int bottomPiece, int bottomPiecePercent, int decimalPosition) {
		return (((bottomPiecePercent != 0) ? 1 : 0) + decimalPosition >=
 2);
	}

	static boolean enableSecondary(int bottomPiece, int bottomPiecePercent, int decimalPosition,
			boolean slot) {
		if (debug)
			checkSecondary(bottomPiece, bottomPiecePercent, decimalPosition, slot);
		return (bottomPiece != 0 && bottomPiecePercent == 0 && decimalPosition >= 2 &&
			slot);
	}

	static boolean enableSubdecks(int bottomPiece, int bottomPiecePercent, int decimalPosition,
			boolean slot, int subbase) {
		if (debug)
			System.out.println("enableSubbase: subbase = " + subbase);
		return (enableSecondary(bottomPiece, bottomPiecePercent, decimalPosition, slot)  &&
			subbase > 0);
	}

	static boolean enableMuseum(int bottomPiece, int bottomPiecePercent, int decimalPosition,
			boolean slot, int subdeck, boolean romanMarkers, int anomaly, int anomalySq) {
		return (enableSecondary(bottomPiece, bottomPiecePercent, decimalPosition, slot)  &&
			subdeck > 0 && romanMarkers && anomaly == 0 && anomalySq == 0);
	}

	static boolean enableRomanMarkers(int anomaly, int anomalySq) {
		return (anomaly == 0 && anomalySq == 0);
	}

	private void checkEnabled() {
		latinMenuItem.setEnabled(enableLatin(romanNumeralsValue.switched(), bottomPieceValue));
		piecePercentRailMenu.setEnabled(enablePiecePercent(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue));
		secondaryRailsMenu.setEnabled(enableSecondary(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue, slotValue));
		subdeckRailsMenu.setEnabled(enableSecondary(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue, slotValue));
		subdecksSeparatedMenuItem.setEnabled(enableSubdecks(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue, slotValue, subbaseValue));
		altSubbeadPlacementMenuItem.setEnabled(enableSubdecks(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue, slotValue, subbaseValue));
		museumMenu.setEnabled(enableMuseum(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue, slotValue, subbaseValue, romanMarkersValue.switched(), anomalyValue, anomalySqValue));
		romanMarkersMenu.setEnabled(enableRomanMarkers(anomalyValue, anomalySqValue));
	}

	class RailChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = railsValue, min;

			railsValue = railChanger.getValue();
			min = ((signValue) ? 1 : 0) +
				((bottomPieceValue == 0) ? 0 : 1) +
				((bottomPiecePercentValue == 0) ? 0 :
				1 + shiftPercentValue) +
				((demoValue) ? AbacusInterface.MIN_DEMO_RAILS :
				AbacusInterface.MIN_RAILS);
			if (railsValue < min) {
				/* Prohibit possible endless recursion */
				if (oldValue >= min)
					railsValue = oldValue;
				else
					railsValue = 2 * shiftPercentValue + 1;
				railChanger.setValue(railsValue);
				return;
			}
			if (railsValue != oldValue) {
				abacus.shiftRails(oldValue, railsValue,
					false, false, false);
			}
		}
	}

	class ComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = modeValue;
			String modeString = (String) ((PartialDisableComboBox)event.getSource()).getSelectedItem();

			modeValue = AbacusInterface.Modes.GENERIC.ordinal();
			for (AbacusInterface.Modes mode : EnumSet.range(AbacusInterface.Modes.CHINESE, AbacusInterface.Modes.MEDIEVAL)) {
				if (modeString.toUpperCase().contains(mode.toString())) {
					modeValue = mode.ordinal();
					try {
						mediumIcon = getIcon(AbacusInterface.ICONS[AbacusInterface.Icons.X32.ordinal()], 32, 32);
						largeIcon = getIcon(AbacusInterface.ICONS[AbacusInterface.Icons.X48.ordinal()], 48, 48);
					} catch (Exception e) {
						break;
					}
					break;
				}
			}
			if (modeValue != AbacusInterface.Modes.GENERIC.ordinal() &&
					oldValue != modeValue) {
				slotValue = (modeValue == AbacusInterface.Modes.ROMAN.ordinal());
				abacus.setMode(modeValue);
				abacus.formatRails(0);
				checkEnabled();
				museumValue = abacus.getMuseum();
				museumMenuItem[museumValue].setSelected(true);
				//make sure digits are not lost when
				//CHINESE -> JAPANESE
			}
		}
	}

	class DisplayBaseChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = displayBaseValue;

			displayBaseValue = displayChanger.getValue();
			if (oldValue != displayBaseValue)
				return;
			abacus.setDisplayBaseAbacus(displayBaseValue);
			if (leeValue && leftAuxAbacus != null) {
				leftAuxAbacus.setDisplayBaseAbacus(displayBaseValue);
			}
			if (leeValue && rightAuxAbacus != null) {
				rightAuxAbacus.setDisplayBaseAbacus(displayBaseValue);
			}
		}
	}

	class AbacusBaseChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = baseValue;

			baseValue = baseChanger.getValue();
			if (oldValue == baseValue)
				return;
			if (bottomNumberValue == oldValue && topSpacesValue == 0) {
				topFactorValue = AbacusCanvas.convertBaseToBottom(baseValue);
				bottomNumberValue = baseValue;
			} else if (topNumberValue * topFactorValue == oldValue) {
				topFactorValue = AbacusCanvas.convertBaseToBottom(baseValue);
				bottomNumberValue = topFactorValue;
				topNumberValue = baseValue / topFactorValue;
			} else if ((topNumberValue + 1) * topFactorValue == oldValue) {
				topFactorValue = AbacusCanvas.convertBaseToBottom(baseValue);
				bottomNumberValue = topFactorValue - 1;
				topNumberValue = baseValue / topFactorValue - 1;
			} else {
				System.err.println("Base adjustment, strange format: topNumber " +
					topNumberValue +
					", topFactor " + topFactorValue +
					", newBase " + baseValue +
					", oldBase " + oldValue);
			}
			abacus.setBaseAbacus(baseValue);
			if (leeValue && leftAuxAbacus != null) {
				leftAuxAbacus.setBaseAbacus(baseValue);
			}
			if (leeValue && rightAuxAbacus != null) {
				rightAuxAbacus.setBaseAbacus(baseValue);
			}
		}
	}

	class DelayChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = delayValue;

			delayValue = inc * delayChanger.getValue();
			if (delayValue < 0) {
				/* Prohibit possible endless recursion */
				if (oldValue > 0)
					delayValue = 0;
				delayChanger.setValue(delayValue / inc);
				return;
			}
			if (delayValue != oldValue) {
				abacus.setDelay(delayValue);
			}
		}
	}

	class RomanNumeralsListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			AbacusInterface.RomanNumeralsFormat oldValue = abacus.getRomanNumerals();
			try {
				AbacusInterface.RomanNumeralsFormat newValue =
						AbacusInterface.RomanNumeralsFormat.valueOf(event.getActionCommand());
				if (newValue != oldValue) {
					romanNumeralsValue = newValue;
					abacus.setRomanNumerals(romanNumeralsValue);
				}
			} catch (IllegalArgumentException e) {
				// ignore
			}
		}
	}

	class PieceListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = pieceChoice(abacus.getTopPiece(), abacus.getBottomPiece());

			try {
				int newValue = AbacusInterface.PieceFormat.valueOf(event.getActionCommand()).ordinal() * 4;

				if (newValue != oldValue) {
					if (newValue <= 4) {
						topPieceValue = 0;
						bottomPieceValue = newValue;
					} else {
						topPieceValue = 2;
						bottomPieceValue = newValue / 2;
					}
					abacus.setPiece(topPieceValue, bottomPieceValue);
					checkEnabled();
				}
			} catch (IllegalArgumentException e) {
				// ignore
			}
		}
	}

	class PiecePercentListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = pieceChoice(abacus.getTopPiecePercent(),
				abacus.getBottomPiecePercent());

			try {
				int newValue = AbacusInterface.PieceFormat.valueOf(event.getActionCommand()).ordinal() * 4;

				if (newValue != oldValue) {
					if (newValue <= 4) {
						topPiecePercentValue = 0;
						bottomPiecePercentValue = newValue;
					} else {
						topPiecePercentValue = 2;
						bottomPiecePercentValue = newValue / 2;
					}
					abacus.setPiecePercent(topPiecePercentValue,
						bottomPiecePercentValue);
					checkEnabled();
				}
			} catch (IllegalArgumentException e) {
				// ignore
			}
		}
	}

	class SubdeckListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = abacus.getSubbase();

			try {
				int newValue = AbacusInterface.PieceFormat.valueOf(event.getActionCommand()).ordinal() * 4;

				if (newValue != oldValue) {
					subbaseValue = newValue;
					abacus.setSubbase(subbaseValue);
					checkEnabled();
				}
			} catch (IllegalArgumentException e) {
				// ignore
			}
		}
	}

	class MuseumListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = abacus.getMuseum();

			try {
				int newValue = Integer.parseInt(event.getActionCommand());

				if (newValue != oldValue) {
					museumValue = newValue;
					abacus.setMuseum(museumValue);
				}
			} catch (NumberFormatException e) {
				// ignore
			}
		}
	}

	class RomanMarkersListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			AbacusInterface.RomanMarkersFormat oldValue = abacus.getRomanMarkers();
			try {
				AbacusInterface.RomanMarkersFormat newValue =
						AbacusInterface.RomanMarkersFormat.valueOf(event.getActionCommand());
				if (newValue != oldValue) {
					romanMarkersValue = newValue;
					abacus.setRomanMarkers(romanMarkersValue);
					museumMenu.setEnabled(enableMuseum(bottomPieceValue, bottomPiecePercentValue, decimalPositionValue, slotValue, subbaseValue, romanMarkersValue.switched(), anomalyValue, anomalySqValue));
				}
			} catch (IllegalArgumentException e) {
				// ignore
			}
		}
	}

	private void forceDemoParams() {
		if (modeValue == AbacusInterface.Modes.GENERIC.ordinal()) {
			topNumberValue = AbacusInterface.DEFAULT_TOP_NUMBER;
			bottomNumberValue = AbacusInterface.DEFAULT_BOTTOM_NUMBER;
			topFactorValue = AbacusInterface.DEFAULT_TOP_FACTOR;
			bottomOrientValue = AbacusInterface.DEFAULT_BOTTOM_ORIENT;
			topSpacesValue = bottomSpacesValue;
			modeValue = AbacusInterface.Modes.CHINESE.ordinal();
			baseValue = AbacusInterface.DEFAULT_BASE;
			displayBaseValue = AbacusInterface.DEFAULT_BASE;
			abacus.setValuesModeAbacus(
				topOrientValue, bottomOrientValue,
				topNumberValue, bottomNumberValue,
				topFactorValue, bottomFactorValue,
				topSpacesValue, bottomSpacesValue,
				baseValue, displayBaseValue, modeValue);
			modeComboBox.setSelectedItem(modeValue);
			displayChanger.setValue(displayBaseValue);
			baseChanger.setValue(baseValue);
			if (leeValue && leftAuxAbacus != null) {
				leftAuxAbacus.setValuesBaseAbacus(baseValue, displayBaseValue);
			}
			if (leeValue && rightAuxAbacus != null) {
				rightAuxAbacus.setValuesBaseAbacus(baseValue, displayBaseValue);
			}
			try {
				mediumIcon = getIcon(AbacusInterface.ICONS_32X32[modeValue], 32, 32);
				largeIcon = getIcon(AbacusInterface.ICONS_48X48[modeValue], 48, 48);
			} catch (Exception e) {
				// ignore
			}
		} else if (baseValue != AbacusInterface.DEFAULT_BASE ||
				displayBaseValue != AbacusInterface.DEFAULT_BASE) {
			baseValue = AbacusInterface.DEFAULT_BASE;
			displayBaseValue = AbacusInterface.DEFAULT_BASE;
			displayChanger.setValue(displayBaseValue);
			baseChanger.setValue(baseValue);
			abacus.setValuesBaseAbacus(baseValue, displayBaseValue);
			if (leeValue && leftAuxAbacus != null) {
				leftAuxAbacus.setValuesBaseAbacus(baseValue, displayBaseValue);
			}
			if (leeValue && rightAuxAbacus != null) {
				rightAuxAbacus.setValuesBaseAbacus(baseValue, displayBaseValue);
			}
		}
	}

	private void forceTeachParams() {
		if (displayBaseValue != baseValue || !signValue ||
				anomalyValue != 0 || anomalySqValue != 0) {
			displayBaseValue = baseValue;
			signValue = false;
			anomalyValue = 0;
			anomalySqValue = 0;
			if (displayChanger != null) {
				displayChanger.setValue(displayBaseValue);
			}
			abacus.setValuesTeachAbacus(displayBaseValue,
				signValue, anomalyValue, anomalySqValue);
			if (leeValue && leftAuxAbacus != null) {
				leftAuxAbacus.setValuesBaseAbacus(baseValue, displayBaseValue);
				leftAuxAbacus.setValuesTeachAbacus(displayBaseValue,
					signValue, anomalyValue, anomalySqValue);
			}
			if (leeValue && rightAuxAbacus != null) {
				rightAuxAbacus.setValuesBaseAbacus(baseValue, displayBaseValue);
				rightAuxAbacus.setValuesTeachAbacus(displayBaseValue,
					signValue, anomalyValue, anomalySqValue);
			}
		}
	}

	public static Document getDoc() {
		return doc;
	}

	public void readFile() {
		try {
			AccessController.doPrivileged((PrivilegedExceptionAction<Object>) () -> {
				try {
					doc = docBuilder.parse(new File(AbacusInterface.XML_FILE));
				} catch (IOException e) {
					e.printStackTrace();
				} catch (SAXParseException err) {
					System.out.println("** Parsing error" + ", line " +
							err.getLineNumber() + ", uri " + err.getSystemId());
					System.out.println(" " + err.getMessage());
				} catch (SAXException e) {
					Exception x1 = e.getException();
					((x1 == null) ? e : x1).printStackTrace();
				}
				return null;
			});
		} catch (AccessControlException | PrivilegedActionException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++)
			if (command.equals(COMMAND_STRINGS[opt]))
				processKey(COMMAND_CHARS[opt]);
	}

	public void callbackAbacusDemo(int aux, int deck, int rail, int number) {
		if (leeValue && aux == AbacusInterface.LEFT_AUX)
			leftAuxAbacus.directMove(deck, rail, number);
		else if (leeValue && aux == AbacusInterface.RIGHT_AUX)
			rightAuxAbacus.directMove(deck, rail, number);
		else if (aux == AbacusInterface.PRIMARY)
			abacus.directMove(deck, rail, number);
	}

	public void callbackAbacus(AbacusCanvas ab, String value) {
		if (leeValue && leftAuxAbacus.equals(ab)) {
			if ((value.equals("0.0") || value.equals("0")) &&
					leftAuxValueTextPrompt.isVisible())
				return;
			leftAuxValueTextField.setText(value);
		} else if (leeValue && rightAuxAbacus.equals(ab)) {
			if ((value.equals("0.0") || value.equals("0")) &&
					rightAuxValueTextPrompt.isVisible())
				return;
			rightAuxValueTextField.setText(value);
		} else {
			if (leeValue && (value.equals("0.0") || value.equals("0")) &&
					valueTextPrompt.isVisible())
				return;
			valueTextField.setText(value);
		}
	}

	public static void callbackAbacus(int aux, int deck, int rail, int number) {
		System.out.println("  <move>");
		System.out.println("    <code>" + aux + " " + deck + " " + rail +
			" " + number + " 4</code>");
		System.out.println("    <text>");
		System.out.println("      Lesson");
		System.out.println();
		System.out.println();
		System.out.println("      Press Space-bar to Continue");
		System.out.println("    </text>");
		System.out.println("  </move>");
	}

	static StringBuilder cleanString(StringBuilder dirty) {
		int i, j = 0;

		for (i = 0; i < dirty.length(); i++) {
			if (dirty.charAt(i) == '[' ||
					dirty.charAt(i) == ']') {
				dirty.setLength(j);
				return dirty;
			} else if (dirty.charAt(i) != ' ' &&
					dirty.charAt(i) != '\t') {
				dirty.setCharAt(j, dirty.charAt(i));
				j++;
			}
		}
		return dirty;
	}

	public void clearAllBeads(int aux) {
		if (aux == 0) {
			valueTextField.setText("0.0");
		} else if (leeValue && aux == AbacusInterface.LEFT_AUX) {
			leftAuxAbacus.clearAllBeads();
		} else if (leeValue && aux == AbacusInterface.RIGHT_AUX) {
			rightAuxAbacus.clearAllBeads();
		}
	}

	public int pieceChoice(int topPiece, int bottomPiece) {
		if (topPiece == 0 && bottomPiece == 0)
			return 0;
		else if (topPiece == 0 && bottomPiece == 4)
			return 4;
		else if (topPiece == 2 && bottomPiece == 4)
			return 8;
		else if (topPiece == 2 && bottomPiece == 6)
			return 12;
		return 0;
	}

	public void callbackAbacus(AbacusCanvas ab, int reason) {
		switch(reason) {
		/*case AbacusInterface.ACTION_SCRIPT:
			break;*/
		case AbacusInterface.ACTION_MOVE:
			break;
		case AbacusInterface.ACTION_CLEAR:
			if (leeValue) {
				leftAuxAbacus.clearAllBeads();
				rightAuxAbacus.clearAllBeads();
				//leftAuxValueTextField.setText("0.0");
				//rightAuxValueTextField.setText("0.0");
			}
			valueTextField.setText("0.0");
			break;
		case AbacusInterface.ACTION_INCREMENT:
			if (!abacus.equals(ab))
				return;
			railsValue++;
			if (railChanger != null) {
				railChanger.setValue(railsValue);
			}
			break;
		case AbacusInterface.ACTION_DECREMENT:
			if (!abacus.equals(ab))
				return;
			railsValue--;
			if (railChanger != null) {
				railChanger.setValue(railsValue);
			}
			break;
		case AbacusInterface.ACTION_DEMO:
			if (!demoValue) {
				abacus.clearAllBeads();
			}
			if (!abacus.equals(ab))
				return;
			if (getTest() || getTeach())
				return;
			forceDemoParams();
			demoValue = !demoValue;
			break;
		case AbacusInterface.ACTION_TEACH:
			if (!abacus.equals(ab))
				return;
			if (getTest() || demoValue)
				return;
			forceTeachParams();
			if (getTeach())
				killTeachDialog();
			else
				createTeachDialog();
			break;
		case AbacusInterface.ACTION_PLACE_INCREMENT:
			if (!abacus.equals(ab))
				return;
			decimalPositionValue++;
			checkEnabled();
			break;
		case AbacusInterface.ACTION_PLACE_DECREMENT:
			if (!abacus.equals(ab))
				return;
			decimalPositionValue--;
			checkEnabled();
			break;
		case AbacusInterface.ACTION_FORMAT:
			if (!abacus.equals(ab))
				return;
			modeValue = AbacusInterface.Modes.values()[modeValue].next().ordinal();
			slotValue = (modeValue == AbacusInterface.Modes.ROMAN.ordinal());
			modeComboBox.setSelectedIndex(modeValue);
			checkEnabled();
			try {
				mediumIcon = getIcon(AbacusInterface.ICONS_32X32[modeValue], 32, 32);
				largeIcon = getIcon(AbacusInterface.ICONS_48X48[modeValue], 48, 48);
			} catch (Exception e) {
				break;
			}
			break;
		case AbacusInterface.ACTION_ROMAN_NUMERALS:
			if (!abacus.equals(ab))
				return;
			romanNumeralsValue = abacus.getRomanNumerals();
			romanNumeralsMenuItem[romanNumeralsValue.ordinal()].setSelected(true);
			latinMenuItem.setEnabled(enableLatin(romanNumeralsValue.switched(), bottomPieceValue));
			break;
		case AbacusInterface.ACTION_ROMAN_MARKERS:
			if (!abacus.equals(ab))
				return;
			romanMarkersValue = abacus.getRomanMarkers();
			romanMarkersMenuItem[romanMarkersValue.ordinal()].setSelected(true);
			break;
		case AbacusInterface.ACTION_LATIN:
			if (!abacus.equals(ab))
				return;
			latinValue = !latinValue;
			latinMenuItem.setSelected(latinValue);
			break;
		case AbacusInterface.ACTION_GROUP:
			if (!abacus.equals(ab))
				return;
			groupValue = !groupValue;
			groupMenuItem.setSelected(groupValue);
			break;
		case AbacusInterface.ACTION_PLACE_ON_RAIL:
			if (!abacus.equals(ab))
				return;
			placeOnRailValue = !placeOnRailValue;
			placeOnRailMenuItem.setSelected(placeOnRailValue);
			break;
		case AbacusInterface.ACTION_VERTICAL:
			if (!abacus.equals(ab))
				return;
			verticalValue = !verticalValue;
			verticalMenuItem.setSelected(verticalValue);
			break;
		case AbacusInterface.ACTION_SIGN:
			if (!abacus.equals(ab))
				return;
			signValue = !signValue;
			signMenuItem.setSelected(signValue);
			break;
		case AbacusInterface.ACTION_PIECE:
			if (!abacus.equals(ab))
				return;
			topPieceValue = abacus.getTopPiece();
			bottomPieceValue = abacus.getBottomPiece();
			pieceMenuItem[pieceChoice(topPieceValue, bottomPieceValue) >> 2].setSelected(true);
			checkEnabled();
			break;
		case AbacusInterface.ACTION_PIECE_PERCENT:
			if (!abacus.equals(ab))
				return;
			topPiecePercentValue = abacus.getTopPiecePercent();
			bottomPiecePercentValue = abacus.getBottomPiecePercent();
			piecePercentMenuItem[pieceChoice(topPiecePercentValue, bottomPiecePercentValue) >> 2].setSelected(true);
			checkEnabled();
			break;
		case AbacusInterface.ACTION_SUBDECK:
			if (!abacus.equals(ab))
				return;
			subbaseValue = abacus.getSubbase();
			checkEnabled();
			break;
		case AbacusInterface.ACTION_SUBDECKS_SEPARATED:
			if (!abacus.equals(ab))
				return;
			subdecksSeparatedValue = !subdecksSeparatedValue;
			subdecksSeparatedMenuItem.setSelected(subdecksSeparatedValue);
			break;
		case AbacusInterface.ACTION_ALT_SUBBEAD_PLACEMENT:
			if (!abacus.equals(ab))
				return;
			altSubbeadPlacementValue = !altSubbeadPlacementValue;
			altSubbeadPlacementMenuItem.setSelected(altSubbeadPlacementValue);
			break;
		case AbacusInterface.ACTION_MUSEUM:
			museumValue = abacus.getMuseum();
			museumMenuItem[museumValue].setSelected(true);
			break;
		case AbacusInterface.ACTION_ANOMALY:
			if (!abacus.equals(ab))
				return;
			anomalyValue = abacus.getAnomaly();
			anomalySqValue = abacus.getAnomalySq();
			checkEnabled();
			romanMarkersValue = abacus.getRomanMarkers();
			romanMarkersMenuItem[romanMarkersValue.ordinal()].setSelected(true);
			break;
		case AbacusInterface.ACTION_RIGHT_TO_LEFT_ADD:
			if (!abacus.equals(ab))
				return;
			rightToLeftAddMenuItem.setSelected(rightToLeftAddValue);
			break;
		case AbacusInterface.ACTION_RIGHT_TO_LEFT_MULT:
			if (!abacus.equals(ab))
				return;
			rightToLeftMultMenuItem.setSelected(rightToLeftMultValue);
			break;
		case AbacusInterface.ACTION_SPEED_UP:
			delayValue -= inc;
			if (delayValue < 0)
				break;
			delayValue -= inc;
			if (delayValue < 0)
				delayValue = 0;
			if (delayChanger != null) {
				delayChanger.setValue(delayValue / inc);
			}
			break;
		case AbacusInterface.ACTION_SLOW_DOWN:
			if (delayValue > MAX_DELAY)
				break;
			delayValue += inc;
			if (delayValue > MAX_DELAY)
				delayValue = MAX_DELAY;
			if (delayChanger != null) {
				delayChanger.setValue(delayValue / inc);
			}
			break;
		default:
			System.out.println("Callback:" + reason);
		}
	}

	public AbacusFrame(String title) {
		/*JPanel info0Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info1Panel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel valuePanel = new JPanel();
		JPanel leftAuxValuePanel = new JPanel();
		JPanel rightAuxValuePanel = new JPanel();
		JPanel basicControlPanel = new JPanel();
		JPanel comboBoxPanel = new JPanel();
		JPanel canvasPanel = new JPanel();
		JPanel leftAuxCanvasPanel = new JPanel();
		JPanel rightAuxCanvasPanel = new JPanel();

		/*// Determine what the default GraphicsDevice can support.
		GraphicsEnvironment ge =
			GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gd = ge.getDefaultScreenDevice();

		boolean isShapedWindowSupported =
			gd.isWindowTranslucencySupported(PERPIXEL_TRANSPARENT);*/

		generator = new Random(System.nanoTime());
		getOpts();
		try {
			mediumIcon = getIcon(AbacusInterface.ICONS_32X32[modeValue], 32, 32);
			largeIcon = getIcon(AbacusInterface.ICONS_48X48[modeValue], 48, 48);
		} catch (Exception e) {
			//
		}
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (modeValue != AbacusInterface.Modes.GENERIC.ordinal())
			slotValue = (modeValue == AbacusInterface.Modes.ROMAN.ordinal());
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu, submenu;
			JMenuItem menuItem;

			// Allows menus to be on top of canvas
			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Controls");
			submenu = new JMenu("Bead Control");
			menuItem = new JMenuItem("Clear");
			submenu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Complement -");
			submenu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Undo");
			submenu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			submenu.add(menuItem);
			menuItem.addActionListener(this);
			menu.add(submenu);
			/* redundant */
			/*submenu = new JMenu("Basic Format");
			menuItem = new JMenuItem("Format");
			submenu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_F);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Increment");
			submenu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_I);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Decrement");
			submenu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_D);
			menuItem.addActionListener(this);
			menu.add(submenu);*/
			submenu = new JMenu("Base Settings");
			/*displayChanger = new Spinner("Display Base",
				displayBaseValue,
				AbacusInterface.MIN_BASE, AbacusInterface.MAX_BASE);
			submenu.add(displayChanger);
			baseChanger = new Spinner("Abacus Base",
				baseValue,
				AbacusInterface.MIN_BASE, AbacusInterface.MAX_BASE);
			submenu.add(baseChanger);*/
			displayChanger = new Slider("Display Base",
				displayBaseValue,
				AbacusInterface.MIN_BASE, AbacusInterface.MAX_BASE,
				126);
			submenu.add(displayChanger);
			baseChanger = new Slider("Abacus Base",
				baseValue,
				AbacusInterface.MIN_BASE, AbacusInterface.MAX_BASE,
				126);
			submenu.add(baseChanger);
			menu.add(submenu);
			submenu = new JMenu("Display Format");
			romanNumeralsMenu = new JMenu("Roman Numerals");
			ButtonGroup romanNumeralsGroup = new ButtonGroup();
			romanNumeralsMenuItem = new JRadioButtonMenuItem[AbacusInterface.RomanNumeralsFormat.values().length];
			for (AbacusInterface.RomanNumeralsFormat format :
					AbacusInterface.RomanNumeralsFormat.values()) {
				romanNumeralsMenuItem[format.ordinal()] =
						new JRadioButtonMenuItem(format.toTitleString());
				romanNumeralsMenu.add(romanNumeralsMenuItem[format.ordinal()]);
				romanNumeralsMenuItem[format.ordinal()].setActionCommand(format.toString());
				romanNumeralsGroup.add(romanNumeralsMenuItem[format.ordinal()]);
				if (romanNumeralsValue == format)
					romanNumeralsMenuItem[format.ordinal()].setSelected(true);
			}
			submenu.add(romanNumeralsMenu);
			romanMarkersMenu = new JMenu("Roman Markerz");
			ButtonGroup romanMarkersGroup = new ButtonGroup();
			romanMarkersMenuItem = new JRadioButtonMenuItem[AbacusInterface.RomanMarkersFormat.values().length];
			for (AbacusInterface.RomanMarkersFormat format :
					AbacusInterface.RomanMarkersFormat.values()) {
				romanMarkersMenuItem[format.ordinal()] =
						new JRadioButtonMenuItem(format.toTitleString());
				romanMarkersMenu.add(romanMarkersMenuItem[format.ordinal()]);
				romanMarkersMenuItem[format.ordinal()].setActionCommand(format.toString());
				romanMarkersGroup.add(romanMarkersMenuItem[format.ordinal()]);
				if (romanMarkersValue == format)
					romanMarkersMenuItem[format.ordinal()].setSelected(true);
			}
			submenu.add(romanMarkersMenu);
			latinMenuItem = new JCheckBoxMenuItem(
				"Latin ~", latinValue);
			submenu.add(latinMenuItem);
			latinMenuItem.addActionListener(this);
			latinMenuItem.setEnabled(enableLatin(romanNumeralsValue.switched(), bottomPieceValue));
			groupMenuItem = new JCheckBoxMenuItem(
				"Group", groupValue);
			submenu.add(groupMenuItem);
			groupMenuItem.setMnemonic(KeyEvent.VK_G);
			groupMenuItem.addActionListener(this);
			menu.add(submenu);
			placeOnRailMenuItem = new JCheckBoxMenuItem(
				"Place On Rail #", placeOnRailValue);
			//placeOnRailMenuItem.setMnemonic(KeyEvent.VK_NUMBER_SIGN);
			submenu.add(placeOnRailMenuItem);
			placeOnRailMenuItem.addActionListener(this);
			menu.add(submenu);
			verticalMenuItem = new JCheckBoxMenuItem(
				"Vertical |", verticalValue);
			submenu.add(verticalMenuItem);
			verticalMenuItem.addActionListener(this);
			menu.add(submenu);
			submenu = new JMenu("Special Rails");
			signMenuItem = new JCheckBoxMenuItem(
				"Sign", signValue);
			submenu.add(signMenuItem);
			signMenuItem.setMnemonic(KeyEvent.VK_S);
			signMenuItem.addActionListener(this);
			pieceRailMenu = new JMenu("Piece");
			ButtonGroup pieceGroup = new ButtonGroup();
			pieceMenuItem = new JRadioButtonMenuItem[AbacusInterface.PieceFormat.values().length];
			for (AbacusInterface.PieceFormat format :
					AbacusInterface.PieceFormat.values()) {
				pieceMenuItem[format.ordinal()] =
						new JRadioButtonMenuItem(format.toTitleString());
				pieceRailMenu.add(pieceMenuItem[format.ordinal()]);
				pieceMenuItem[format.ordinal()].setActionCommand(format.toString());
				pieceGroup.add(pieceMenuItem[format.ordinal()]);
				if (pieceChoice(topPieceValue, bottomPieceValue) == 4 * format.ordinal())
					pieceMenuItem[format.ordinal()].setSelected(true);
			}
			submenu.add(pieceRailMenu);
			piecePercentRailMenu = new JMenu("Piece Percent");
			ButtonGroup piecePercentGroup = new ButtonGroup();
			piecePercentMenuItem = new JRadioButtonMenuItem[AbacusInterface.PieceFormat.values().length];
			for (AbacusInterface.PieceFormat format :
					AbacusInterface.PieceFormat.values()) {
				piecePercentMenuItem[format.ordinal()] =
						new JRadioButtonMenuItem(format.toTitleString());
				piecePercentRailMenu.add(piecePercentMenuItem[format.ordinal()]);
				piecePercentMenuItem[format.ordinal()].setActionCommand(format.toString());
				piecePercentGroup.add(piecePercentMenuItem[format.ordinal()]);
				if (pieceChoice(topPiecePercentValue, bottomPiecePercentValue) == 4 * format.ordinal())
					piecePercentMenuItem[format.ordinal()].setSelected(true);
			}
			submenu.add(piecePercentRailMenu);
			//submenu.addSeparator();
			secondaryRailsMenu = new JMenu("Secondary Rails");
			subdeckRailsMenu = new JMenu("Subdeck");
			ButtonGroup subdeckGroup = new ButtonGroup();
			subdeckMenuItem = new JRadioButtonMenuItem[AbacusInterface.PieceFormat.values().length];
			for (AbacusInterface.PieceFormat format :
					AbacusInterface.PieceFormat.values()) {
				subdeckMenuItem[format.ordinal()] =
						new JRadioButtonMenuItem(format.toTitleString());
				subdeckRailsMenu.add(subdeckMenuItem[format.ordinal()]);
				subdeckMenuItem[format.ordinal()].setActionCommand(format.toString());
				subdeckGroup.add(subdeckMenuItem[format.ordinal()]);
				if (subbaseValue == 4 * format.ordinal())
					subdeckMenuItem[format.ordinal()].setSelected(true);
			}
			secondaryRailsMenu.add(subdeckRailsMenu);
			subdecksSeparatedMenuItem = new JCheckBoxMenuItem(
				"Subdecks Separated", subdecksSeparatedValue);
			secondaryRailsMenu.add(subdecksSeparatedMenuItem);
			subdecksSeparatedMenuItem.setMnemonic(KeyEvent.VK_K);
			subdecksSeparatedMenuItem.addActionListener(this);
			altSubbeadPlacementMenuItem = new JCheckBoxMenuItem(
				"Alt Subbead Placement", altSubbeadPlacementValue);
			secondaryRailsMenu.add(altSubbeadPlacementMenuItem);
			altSubbeadPlacementMenuItem.setMnemonic(KeyEvent.VK_L);
			altSubbeadPlacementMenuItem.addActionListener(this);
			museumMenu = new JMenu("Museum");
			ButtonGroup museumGroup = new ButtonGroup();
			museumMenuItem = new JRadioButtonMenuItem[AbacusInterface.MUSEUM_COUNTRY.length];
			for (int i = 0; i < AbacusInterface.MUSEUM_COUNTRY.length; i++) {
				museumMenuItem[i] =
						new JRadioButtonMenuItem(AbacusInterface.MUSEUM_COUNTRY[i]);
				museumMenu.add(museumMenuItem[i]);
				museumMenuItem[i].setActionCommand(Integer.toString(i));
				museumGroup.add(museumMenuItem[i]);
			}
			secondaryRailsMenu.add(museumMenu);
			submenu.add(secondaryRailsMenu);
			/*submenu.addSeparator();
			anomalyMenuItem = new JCheckBoxMenuItem(
				"Anomaly", setAnomaly(anomalyValue, anomalySqValue));
			submenu.add(anomalyMenuItem);
			anomalyMenuItem.setMnemonic(KeyEvent.VK_Y);
			anomalyMenuItem.addActionListener(this);*/
			menu.add(submenu);
			submenu = new JMenu("Teach Options");
			rightToLeftAddMenuItem = new JCheckBoxMenuItem(
				"Right To Left Add +", rightToLeftAddValue);
			submenu.add(rightToLeftAddMenuItem);
			rightToLeftAddMenuItem.addActionListener(this);
			rightToLeftMultMenuItem = new JCheckBoxMenuItem(
				"Right To Left Mult *", rightToLeftMultValue);
			submenu.add(rightToLeftMultMenuItem);
			rightToLeftMultMenuItem.addActionListener(this);
			menu.add(submenu);
			submenu = new JMenu("Effects");
			soundMenuItem = new JCheckBoxMenuItem("Sound @", sound);
			submenu.add(soundMenuItem);
			/*delayChanger = new Spinner("msec Delay",
				delayValue,
				0, MAX_DELAY,
				inc);*/
			delayChanger = new Slider("msec Delay",
				delayValue / inc, 0, MAX_DELAY / inc, 126);
			delayChanger.setIncrement(inc);
			submenu.add(delayChanger);
			soundMenuItem.addActionListener(this);
			menu.add(submenu);
			menuBar.add(menu);
			menu = new JMenu("Learn");
			menuItem = new JMenuItem("Demo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_O);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Teach $");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			leftPanel.add(new JLabel("Move bead"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Clear"));
			leftPanel.add(getIcon("/icons/mouse-l.png", 16, 16));
			rightPanel.add(getIcon("/icons/mouse-r.png", 16, 16));
			info0Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info0Panel.add(leftPanel);
			info0Panel.add(rightPanel);*/
			railChanger = new Spinner("Rails",
				railsValue, AbacusInterface.MIN_RAILS);
			/*railChanger = new Slider("Abacus Size",
				railsValue, AbacusInterface.MIN_RAILS,
				((railsValue > MAX_RAILS) ? railsValue : MAX_RAILS),
				126);*/
			basicControlPanel.setLayout(new FlowLayout(FlowLayout.LEFT,
				15, 5));
			comboBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT,
				5, 5));
			comboBoxPanel.add(new JLabel("Format:"));
			for (AbacusInterface.Modes mode : AbacusInterface.Modes.values()) {
				modeComboBox.addItem(AbacusInterface.titleName(mode.toString()) + " " +
					AbacusInterface.FORMAT_STRINGS[mode.ordinal()],
					(mode != AbacusInterface.Modes.GENERIC));
			}
			modeComboBox.setLightWeightPopupEnabled(false);
			modeComboBox.setSelectedItem(AbacusInterface.titleName(
					AbacusInterface.Modes.values()[modeValue].toString()) +
					" " + AbacusInterface.FORMAT_STRINGS[modeValue]);
			comboBoxPanel.add(modeComboBox);
			basicControlPanel.add(railChanger);
			basicControlPanel.add(comboBoxPanel);
		}
		if (leeValue) {
			leftAuxValuePanel.setLayout(new FlowLayout(FlowLayout.LEFT,
				0, 0));
			rightAuxValuePanel.setLayout(new FlowLayout(FlowLayout.LEFT,
				0, 0));
			//set as html table if using JEditorPane
			leftAuxValueTextField.setColumns(valueTextLength *
				leftAuxRailsValue / totalAuxRailsValue + 1);
			rightAuxValueTextField.setColumns(valueTextLength *
				rightAuxRailsValue / totalAuxRailsValue + 1);
			leftAuxValuePanel.add(leftAuxValueTextField);
			rightAuxValuePanel.add(rightAuxValueTextField);
			if (testValue == 0) {
				info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
				info1Panel.add(leftAuxValuePanel);
				info1Panel.add(rightAuxValuePanel);
			}
			leftAuxValueTextPrompt = new TextPrompt("Left Auxiliary",
				leftAuxValueTextField);
			leftAuxValueTextPrompt.changeStyle(Font.ITALIC);
			leftAuxValueTextPrompt.setForeground(leftAuxBeadColor.darker().darker());
			rightAuxValueTextPrompt = new TextPrompt("Right Auxiliary",
				rightAuxValueTextField);
			rightAuxValueTextPrompt.changeStyle(Font.ITALIC);
			rightAuxValueTextPrompt.setForeground(rightAuxBeadColor.darker().darker());
			valueTextPrompt = new TextPrompt("Primary", valueTextField);
			valueTextPrompt.changeStyle(Font.ITALIC + Font.BOLD);
			valueTextPrompt.setForeground(primaryBeadColor.darker().darker());
		} else {
			valueTextPrompt = new TextPrompt("0", valueTextField);
			valueTextPrompt.changeStyle(Font.BOLD);
		}
		if (!controlValue)
			info1Panel.setBackground(bgColor);
		if (testValue == 0) {
			valuePanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			//set as html table if using JEditorPane
			valueTextField.setColumns(valueTextLength);
			//valueTextField.setEditable(false);
			valuePanel.add(valueTextField);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			if (!controlValue)
				info2Panel.setBackground(bgColor);
			info2Panel.add(valuePanel);
		}
		topPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		/*topPanel.add(info0Panel);*/
		if (controlValue) {
			topPanel.add(basicControlPanel);
		} else
			topPanel.setBackground(bgColor);
		if (testValue == 0 && leeValue) {
			topPanel.add(info1Panel);
		}
		if (testValue == 0)
			topPanel.add(info2Panel);
		abacus = new Abacus(this,
			fgColor, bgColor, borderColor,
			primaryBeadColor, secondaryBeadColor, highlightBeadColor, primaryMarkerColor,
			primaryRailColor, secondaryRailColor, highlightRailColor, lineRailColor,
			delayValue, railsValue, verticalValue,
			colorSchemeValue, slotValue,
			diamondValue, railIndexValue,
			topOrientValue, bottomOrientValue,
			topNumberValue, bottomNumberValue,
			topFactorValue, bottomFactorValue,
			topSpacesValue, bottomSpacesValue,
			topPieceValue, bottomPieceValue,
			topPiecePercentValue, bottomPiecePercentValue,
			shiftPercentValue, subdeckValue, subbeadValue,
			signValue, decimalPositionValue,
			groupValue, groupSizeValue, placeOnRailValue,
			decimalCommaValue, baseValue, subbaseValue,
			anomalyValue, shiftAnomalyValue,
			anomalySqValue, shiftAnomalySqValue,
			displayBaseValue, (pressOffsetValue) ? 1 : 0,
			romanNumeralsValue, latinValue,
			romanMarkersValue,
			modeValue, museumValue,
			subdecksSeparatedValue,
			altSubbeadPlacementValue);
		if (leeValue) {
		int auxModeValue = AbacusInterface.Modes.JAPANESE.ordinal();
		leftAuxAbacus = new Abacus(this,
			fgColor, bgColor, borderColor,
			leftAuxBeadColor, secondaryBeadColor, highlightBeadColor, leftAuxMarkerColor,
			primaryRailColor, secondaryRailColor, highlightRailColor, lineRailColor,
			delayValue, leftAuxRailsValue, verticalValue,
			colorSchemeValue, slotValue,
			diamondValue, railIndexValue,
			topOrientValue, bottomOrientValue,
			topNumberValue, bottomNumberValue,
			topFactorValue, bottomFactorValue,
			topSpacesValue, bottomSpacesValue,
			topPieceValue, bottomPieceValue,
			topPiecePercentValue, bottomPiecePercentValue,
			shiftPercentValue, subdeckValue, subbeadValue,
			signValue, decimalPositionValue,
			groupValue, groupSizeValue, placeOnRailValue,
			decimalCommaValue, baseValue, subbaseValue,
			anomalyValue, shiftAnomalyValue,
			anomalySqValue, shiftAnomalySqValue,
			displayBaseValue, (pressOffsetValue) ? 1 : 0,
			romanNumeralsValue, latinValue,
			romanMarkersValue,
			auxModeValue, museumValue,
			subdecksSeparatedValue,
			altSubbeadPlacementValue);
		rightAuxAbacus = new Abacus(this,
			fgColor, bgColor, borderColor,
			rightAuxBeadColor, secondaryBeadColor, highlightBeadColor, rightAuxMarkerColor,
			primaryRailColor, secondaryRailColor, highlightRailColor, lineRailColor,
			delayValue, rightAuxRailsValue, verticalValue,
			colorSchemeValue, slotValue,
			diamondValue,railIndexValue,
			topOrientValue, bottomOrientValue,
			topNumberValue, bottomNumberValue,
			topFactorValue, bottomFactorValue,
			topSpacesValue, bottomSpacesValue,
			topPieceValue, bottomPieceValue,
			topPiecePercentValue, bottomPiecePercentValue,
			shiftPercentValue, subdeckValue, subbeadValue,
			signValue, decimalPositionValue,
			groupValue, groupSizeValue, placeOnRailValue,
			decimalCommaValue, baseValue, subbaseValue,
			anomalyValue, shiftAnomalyValue,
			anomalySqValue, shiftAnomalySqValue,
			displayBaseValue, (pressOffsetValue) ? 1 : 0,
			romanNumeralsValue, latinValue,
			romanMarkersValue,
			auxModeValue, museumValue,
			subdecksSeparatedValue,
			altSubbeadPlacementValue);
		}
		startPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		if (!controlValue)
			startPanel.setBackground(bgColor);
		startPanel.add(topPanel);
		secondaryPanel.setLayout(new BorderLayout());
		secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		canvasPanel.setBackground(bgColor);
		if (leeValue) {
			leftAuxCanvasPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			rightAuxCanvasPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			canvasPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			leftAuxCanvasPanel.add(leftAuxAbacus);
			rightAuxCanvasPanel.add(rightAuxAbacus);
			canvasPanel.add(abacus);
			JPanel leePanel = new JPanel();
			JPanel extraPanel = new JPanel();
			leePanel.setLayout(new GridBagLayout());
			GridBagConstraints c = new GridBagConstraints();
			c.anchor = GridBagConstraints.FIRST_LINE_START;
			c.weightx = 0.5;
			c.gridx = 0;
			c.gridy = 0;
			leePanel.add(leftAuxCanvasPanel, c);
			c.anchor = GridBagConstraints.FIRST_LINE_END;
			c.gridx = 1;
			c.gridy = 0;
			leePanel.add(rightAuxCanvasPanel, c);
			c.weightx = 0.0;
			c.gridwidth = 2;
			c.anchor = GridBagConstraints.PAGE_START;
			c.gridx = 0;
			c.gridy = 1;
			leePanel.add(canvasPanel, c);
			extraPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			extraPanel.add(leePanel);
			secondaryPanel.add(extraPanel, BorderLayout.CENTER);
			extraPanel.setBackground(bgColor);
		} else {
			canvasPanel.setBackground(bgColor);
			canvasPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			canvasPanel.add(abacus);
			secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		}
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			railChanger.addChangeListener(new RailChangeListener());
			modeComboBox.addActionListener(new ComboBoxActionListener());
			displayChanger.addChangeListener(
				new DisplayBaseChangeListener());
			baseChanger.addChangeListener(
				new AbacusBaseChangeListener());
			for (AbacusInterface.RomanNumeralsFormat format :
					AbacusInterface.RomanNumeralsFormat.values()) {
				romanNumeralsMenuItem[format.ordinal()].addActionListener(new RomanNumeralsListener());
			}
			delayChanger.addChangeListener( new DelayChangeListener());
			for (AbacusInterface.PieceFormat format :
					AbacusInterface.PieceFormat.values()) {
				pieceMenuItem[format.ordinal()].addActionListener(new PieceListener());
			}
			for (AbacusInterface.PieceFormat format :
					AbacusInterface.PieceFormat.values()) {
				piecePercentMenuItem[format.ordinal()].addActionListener(new PiecePercentListener());
			}
			for (AbacusInterface.PieceFormat format :
					AbacusInterface.PieceFormat.values()) {
				subdeckMenuItem[format.ordinal()].addActionListener(new SubdeckListener());
			}
			for (int i = 0; i < AbacusInterface.Museum.values().length; i++)
				museumMenuItem[i].addActionListener(new MuseumListener());
			for (AbacusInterface.RomanMarkersFormat format :
					AbacusInterface.RomanMarkersFormat.values()) {
				romanMarkersMenuItem[format.ordinal()].addActionListener(new RomanMarkersListener());
			}
		}
		if (leeValue) {
			leftAuxValueTextField.addActionListener((ActionEvent event) -> {
				StringBuilder mathBuilder = new
								StringBuilder(leftAuxValueTextField.getText());
				if (mathBuilder.length() >= 0)
					leftAuxAbacusCalc.calculateAndDisplay(leftAuxAbacus,
							cleanString(mathBuilder).toString(),
							AbacusInterface.LEFT_AUX);
				else
					leftAuxAbacus.clearAllBeads();
			});
			leftAuxCanvasPanel.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseEntered(MouseEvent event) {
					leftAuxAbacus.requestFocus();
				}
				@Override
				public void mouseExited(MouseEvent event) {
					leftAuxAbacus.requestFocus();
				}
				@Override
				public void mousePressed(MouseEvent event) {
					leftAuxAbacus.requestFocus();
				}
			});
			rightAuxValueTextField.addActionListener((ActionEvent event) -> {
				StringBuilder mathBuilder = new
								StringBuilder(rightAuxValueTextField.getText());

				if (mathBuilder.length() >= 0)
					rightAuxAbacusCalc.calculateAndDisplay(rightAuxAbacus,
							cleanString(mathBuilder).toString(),
							AbacusInterface.RIGHT_AUX);
				else
					rightAuxAbacus.clearAllBeads();
			});
			rightAuxCanvasPanel.addMouseListener(new MouseAdapter() {
				@Override
				public void mouseEntered(MouseEvent event) {
					rightAuxAbacus.requestFocus();
				}
				@Override
				public void mouseExited(MouseEvent event) {
					rightAuxAbacus.requestFocus();
				}
				@Override
				public void mousePressed(MouseEvent event) {
					rightAuxAbacus.requestFocus();
				}
			});
		}
		valueTextField.addActionListener((ActionEvent event) -> {
			StringBuilder mathBuilder = new
							StringBuilder(valueTextField.getText());

			if (mathBuilder.length() >= 0)
				abacusCalc.calculateAndDisplay(abacus,
						cleanString(mathBuilder).toString(),
						AbacusInterface.PRIMARY);
			else
				abacus.clearAllBeads();
		});
		canvasPanel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				abacus.requestFocus();
			}
			@Override
			public void mouseExited(MouseEvent event) {
				abacus.requestFocus();
			}
			@Override
			public void mousePressed(MouseEvent event) {
				abacus.requestFocus();
			}
		});
		setTitle(title);
		/* abacus.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		if (leeValue) {
			int newWidth = leftAuxRailsValue * widthValue /
				totalAuxRailsValue;
			int newHeight = heightValue / 4;

			leftAuxAbacus.setSize(newWidth, newHeight);
			rightAuxAbacus.setSize(widthValue - newWidth,
				newHeight);
			abacus.setSize(widthValue, heightValue - newHeight);
		} else {
			abacus.setSize(widthValue, heightValue);
		}
		Font boldFont = valueTextField.getFont().deriveFont(Font.BOLD);
		valueTextField.setFont(boldFont);
		try {
			docBuilderFactory =
				DocumentBuilderFactory.newInstance();
			docBuilder = docBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		}
		/* ap let
		if (testValue != 0) {
			forceDemoParams();
			abacus.checkBeads();
			abacus.resetBeads();
			AbacusTest abacusTest = new AbacusTest(this, abacus);
			abacusTest.getTest();
		}*/
		if (controlValue) {
			// decimalPosition is hard to figure out beforehand
			checkEnabled();
			museumMenuItem[museumValue].setSelected(true);
		}
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : AbacusInterface.ICONS) {
			URL url = AbacusFrame.class.getResource("/" + icon);
			if (url != null) {
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
			}
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-primaryBeadColor={int}] [-leftAuxBeadColor={int}]" + NL +
"\t[-rightAuxBeadColor={int}] [-secondaryBeadColor={int}]" + NL +
"\t[-highlightBeadColor={int}] [-primaryMarkerColor={int}]" + NL +
"\t[-leftAuxMarkerColor={int}] [-rightAuxMarkerColor={int}]" + NL +
"\t[-primaryRailColor={int}] [-secondaryRailColor={int}]" + NL +
"\t[-highlightRailColor={int}] [-lineRailColor={int}]" + NL +
"\t[-borderColor={int}] [-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-delay={int}] [-script={bool}]" + NL +
"\t[-rightToLeftAdd={bool}] [-rightToLeftMult={bool}]" + NL +
"\t[-control={bool}] [-lee={bool}] [-rails={int}]" + NL +
"\t[-leftAuxRails={int}] [-rightAuxRails={int}] [-vertical={bool}]" + NL +
"\t[-colorScheme={int}] [-slot={bool}] [-diamond={bool}]" + NL +
"\t[-railIndex={int}] [-topOrient={bool}] [-bottomOrient={bool}]" + NL +
"\t[-topNumber={int}] [-bottomNumber={int}] [-topFactor={int}]" + NL +
"\t[-bottomFactor={int}] [-topSpaces={int}] [-bottomSpaces={int}]" + NL +
"\t[-topPiece={int}] [-bottomPiece={int}] [-topPiecePercent={int}]" + NL +
"\t[-bottomPiecePercent={int}] [-shiftPercent={int}]" + NL +
"\t[-subdeck={int}] [-subbead={int}] [-sign={bool}]" + NL +
"\t[-decimalPosition={int}] [-group={bool}] [-groupSize={int}]" + NL +
"\t[-placeOnRail={bool}] [-decimalComma={bool}] [-base={int}]" + NL +
"\t[-subbase={int}] [-subdecksSeparated={bool}]" + NL +
"\t[-altSubbeadPlacement={bool}] [-anomaly={int}] [-shiftAnomaly={bool}]" + NL +
"\t[-anomalySq={int}] [-shiftAnomalySq={bool}] [-displayBase={int}]" + NL +
"\t[-pressOffset={bool}] [-romanNumerals={none|ancient|modern}]" + NL +
"\t[-latin={bool}] [-romanMarkers={none|ancient|late|alt|modern}]" + NL +
"\t[-format={chinese|japanese|korean|russian|danish|roman|medieval|generic}]" + NL +
"\t[-museum={it|uk|fr}]");
			return;
		}
		AbacusFrame frame = new AbacusFrame(AbacusInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.abacus.requestFocusInWindow();
		frame.setLocation(AbacusFrame.xValue, AbacusFrame.yValue);
		frame.setVisible(true);
		if (frame.testValue != 0) {
			frame.forceDemoParams();
			frame.abacus.checkBeads();
			frame.abacus.resetBeads();
			AbacusTest abacusTest = new AbacusTest(frame, frame.abacus);
			abacusTest.getTest();
		}
	}
}
